import argparse
import os
import sys
import logging

from backuprotation import PROJECTNAME, VERSION

logger = logging.getLogger(__name__)


def parse_args(args):
    parser = argparse.ArgumentParser(
        prog="backuprotation",
        description="Rotates backups for you.",
        epilog=f"{PROJECTNAME} {VERSION}"
    )

    parser.add_argument(
        "path",
        help="The path of the backup directory."
    )

    parser.add_argument(
        "-n", "--number",
        action="store",
        default=10,
        type=int,
        help="The number of backups to keep."
    )

    parser.add_argument(
        "-f", "--files",
        action="store_true",
        help=f"If this option is selected {PROJECTNAME} looks for files "
             f"instead of directories to rotate."
    )

    parser.add_argument(
        "-fd", "--files-and-directories",
        action="store_true",
        help=f"If this option is selected {PROJECTNAME} looks both for files "
             f"and directories to rotate."
    )

    parser.add_argument(
        "-d", "--dry-run",
        action="store_true",
        help="dry-run: do not delete anything"
    )

    return parser.parse_args(args)


def discover(path, dirs=True, files=False):
    discovered = list()
    for f in os.listdir(os.path.abspath(path)):
        f = os.path.join(path, f)
        if dirs and os.path.isdir(f):
            discovered.append(f)
        if files and os.path.isfile(f):
            discovered.append(f)

    return sorted(discovered, key=lambda d: os.path.getmtime(d))


def rotate(discovered, number, dry_run):
    if len(discovered) <= number:
        return

    scheduled = discovered[:len(discovered) - number]

    for element in scheduled:
        if os.path.isfile(element):
            logger.info(f"deleting file {element}")
            if not dry_run:
                os.remove(element)

        else:
            logger.info(f"deleting directory {element}")
            if not dry_run:
                os.rmdir(element)


def main():
    args = parse_args(sys.argv[1:])

    if args.files:
        dirs = False
        files = True
    elif args.files_and_directories:
        dirs = True
        files = True
    else:
        dirs = True
        files = False

    discovered = discover(args.path, dirs, files)

    rotate(discovered, args.number, args.dry_run)



from backuprotation.main import parse_args, discover, rotate
import pytest
import tempfile
import random
import string
import os
import time


@pytest.fixture
def create_testing_dir():
    tmpdir = tempfile.TemporaryDirectory()

    dirs = list()
    files = list()

    dirnum = random.randint(1, 10)
    filenum = random.randint(1, 10)

    for _ in range(dirnum):
        name = "".join(random.choices(string.ascii_lowercase, k=8))

        while os.path.exists(name):
            name = "".join(random.choices(string.ascii_lowercase, k=8))

        os.makedirs(os.path.join(tmpdir.name, name))
        dirs.append(name)

    for _ in range(filenum):
        name = "".join(random.choices(string.ascii_lowercase, k=8))

        while os.path.exists(name):
            name = "".join(random.choices(string.ascii_lowercase, k=8))

        open(os.path.join(tmpdir.name, name), "w").close()
        files.append(name)

    return tmpdir, dirs, files


@pytest.fixture
def create_timed_dir():
    tmpdir = tempfile.TemporaryDirectory()

    dirs = list()

    dirnum = 30

    for _ in range(dirnum):
        name = "".join(random.choices(string.ascii_lowercase, k=8))

        while os.path.exists(name):
            name = "".join(random.choices(string.ascii_lowercase, k=8))

        time.sleep(0.1)
        os.makedirs(os.path.join(tmpdir.name, name))
        dirs.append(name)

    return tmpdir, dirs


class TestArgumentParsing:
    def test_required_path(self):
        with pytest.raises(SystemExit):
            args = parse_args([])

    def test_n_type(self):
        with pytest.raises(SystemExit):
            args = parse_args(["./path", "-n", "a"])


class TestDiscovery:
    def test_dir_discovery(self, create_testing_dir):
        tmpdir, dirs, files = create_testing_dir

        discovered = discover(tmpdir.name, True, False)

        assert set(dirs) == set(list(map(os.path.basename, discovered)))

    def test_file_discovery(self, create_testing_dir):
        tmpdir, dirs, files = create_testing_dir

        discovered = discover(tmpdir.name, False, True)

        assert set(files) == set(list(map(os.path.basename, discovered)))

    def test_dir_file_discovery(self, create_testing_dir):
        tmpdir, dirs, files = create_testing_dir

        discovered = discover(tmpdir.name, True, True)

        combined = dirs + files

        assert set(combined) == set(list(map(os.path.basename, discovered)))

    def test_ordering_discovery(self):
        tmpdir = tempfile.TemporaryDirectory()

        dirs = list()

        dirnum = random.randint(1, 10)

        for _ in range(dirnum):
            name = "".join(random.choices(string.ascii_lowercase, k=8))

            while os.path.exists(name):
                name = "".join(random.choices(string.ascii_lowercase, k=8))

            time.sleep(0.1)                             # testing for sorting
            os.makedirs(os.path.join(tmpdir.name, name))
            dirs.append(name)

        discovered = discover(tmpdir.name, True, False)

        assert dirs == list(map(os.path.basename, discovered))


class TestRotation:
    def test_dry_run(self, create_timed_dir):
        tmpdir, dirs = create_timed_dir

        discovered = discover(tmpdir.name, True, False)

        rotate(discovered, 10, True)

        after = os.listdir(tmpdir.name)

        assert set(dirs) == set(after)

    def test_deletion(self, create_timed_dir):
        tmpdir, dirs = create_timed_dir

        n = 10

        discovered = discover(tmpdir.name, True, False)

        rotate(discovered, n, False)

        after = os.listdir(tmpdir.name)

        assert set(dirs[-n:]) == set(after)
